/*
 * Copyright (C) 2009 Carl Worth
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/ .
 *
 * Author: Carl Worth <cworth@cworth.org>
 */

#include <stdint.h>
#include <stdlib.h>
#include <assert.h>

#include <glib.h>

#include "dvonn-board.h"

/* Just a diagram to help start thinking about data structures:
 *
 * Played/drawn as:
 *
 *   1 ○ ○ ○ ○ ○ ○ ○ ○ ○
 *  2 ○ ● ● ○ ○ ○ ○ ○ ○ ○
 * 3 ○ ● ○ ● ○ ○ ○ ○ ○ ○ ○
 *  4 ○ ● ● ○ ○ ○ ○ ○ ○ ○ K
 *   5 ○ ○ ○ ○ ○ ○ ○ ○ ○ J
 *      A B C D E F G H I
 *
 * Stored as:
 *
 * 1 x x ○ ○ ○ ○ ○ ○ ○ ○ ○
 * 2 x ○ ● ● ○ ○ ○ ○ ○ ○ ○
 * 3 ○ ● ○ ● ○ ○ ○ ○ ○ ○ ○
 * 4 ○ ● ● ○ ○ ○ ○ ○ ○ ○ x
 * 5 ○ ○ ○ ○ ○ ○ ○ ○ ○ x x
 *   A B C D E F G H I J K
 *
 * With connections as:
 *
 *         C2  D2
 *          | /
 *  B3 <-> C3 <-> D3
 *        / |
 *      B2 C4
 */

void
dvonn_board_init (dvonn_board_t *board)
{
    int x, y;

    for (x = 0; x < DVONN_BOARD_X_SIZE; x++) {
	for (y = 0; y < DVONN_BOARD_Y_SIZE; y++) {
	    board->cells[x][y].type = DVONN_CELL_EMPTY;
	    board->cells[x][y].height = 0;
	    board->cells[x][y].contains_red = FALSE;
	}
    }

    /* Cut out the unplayable "corners" of the board. */
    for (x = 0; x < DVONN_BOARD_Y_SIZE / 2; x++) {
	for (y = 0; y < (DVONN_BOARD_Y_SIZE / 2) - x; y++) {
	    board->cells[x][y].type = DVONN_CELL_INVALID;
	    board->cells
		[DVONN_BOARD_X_SIZE-1-x]
		[DVONN_BOARD_Y_SIZE-1-y].type = DVONN_CELL_INVALID;
	}
    }

    board->phase = DVONN_PHASE_PLACEMENT;
    board->player = DVONN_PLAYER_WHITE;
    board->moves = 0;
    board->score[DVONN_PLAYER_BLACK] = 0;
    board->score[DVONN_PLAYER_WHITE] = 0;
}

dvonn_bool_t
dvonn_board_cell_occupied (dvonn_board_t *board,
			   int x, int y)
{
    if (x < 0 || x >= DVONN_BOARD_X_SIZE ||
	y < 0 || y >= DVONN_BOARD_Y_SIZE)
    {
	return FALSE;
    }

    if (board->cells[x][y].type == DVONN_CELL_INVALID ||
	board->cells[x][y].type == DVONN_CELL_EMPTY)
    {
	return FALSE;
    }

    return TRUE;
}

dvonn_bool_t
dvonn_board_cell_owned_by (dvonn_board_t *board,
			   int x, int y,
			   dvonn_player_t player)
{
    if (! dvonn_board_cell_occupied (board, x, y))
	return FALSE;

    /* Cast here to avoid compiler warning about mixing enum types in
     * a comparison. */
    return board->cells[x][y].type == (dvonn_cell_type_t) player;
}

dvonn_bool_t
dvonn_board_cell_surrounded (dvonn_board_t *board,
			     int x, int y)
{
    int surround_count;

    surround_count =
	dvonn_board_cell_occupied (board, x - 1, y) +
	dvonn_board_cell_occupied (board, x + 1, y) +
	dvonn_board_cell_occupied (board, x, y - 1) +
	dvonn_board_cell_occupied (board, x, y + 1) +
	dvonn_board_cell_occupied (board, x + 1, y - 1) +
	dvonn_board_cell_occupied (board, x - 1, y + 1);

    return (surround_count == 6);
}

static dvonn_bool_t
dvonn_board_move_legal (dvonn_board_t *board,
			int x1, int y1,
			int x2, int y2,
			char **error)
{
    int distance;

    if (x1 < 0 || x1 >= DVONN_BOARD_X_SIZE ||
	y1 < 0 || y1 >= DVONN_BOARD_Y_SIZE ||
	x2 < 0 || x2 >= DVONN_BOARD_X_SIZE ||
	y2 < 0 || y2 >= DVONN_BOARD_Y_SIZE)
    {
	*error = "Invalid coordinates (not on board)";
	return FALSE;
    }

    if (board->cells[x1][y1].type == DVONN_CELL_INVALID ||
	board->cells[x2][y2].type == DVONN_CELL_INVALID)
    {
	*error = "Not a valid board space";
	return FALSE;
    }

    if (board->cells[x1][y1].type == DVONN_CELL_EMPTY) {
	*error = "There are no pieces there to move";
	return FALSE;
    }

    if (dvonn_board_cell_surrounded (board, x1, y1)) {
	*error = "You cannot move a piece that is surrounded";
	return FALSE;
    }

    if (! dvonn_board_cell_owned_by (board, x1, y1, board->player)) {
	*error = "You cannot move your opponent's stack";
	return FALSE;
    }

    if (board->cells[x2][y2].type == DVONN_CELL_EMPTY) {
	*error = "You cannot move to an empty space";
	return FALSE;
    }

    /* Move must be in a straight line, which in our array is
     * horizontal, vertical, or one of the diagonals. */
    if (x2 == x1)
	distance = abs (y2 - y1);
    else if (y2 == y1)
	distance = abs (x2 - x1);
    else if ((x2 - x1) == - (y2 - y1))
	distance = abs (x2 - x1);
    else {
	*error = "Move is not in a straight line";
	return FALSE;
    }

    if (distance != board->cells[x1][y1].height) {
	*error = "Move distance is not the same as stack height";
	return FALSE;
    }

    return TRUE;
}

static dvonn_bool_t
dvonn_board_player_has_legal_move (dvonn_board_t *board)
{
    int x, y, height;
    char *error;

    for (x = 0; x < DVONN_BOARD_X_SIZE; x++) {
	for (y = 0; y < DVONN_BOARD_X_SIZE; y++) {
	    if (! dvonn_board_cell_occupied (board, x, y))
		continue;
	    if (! (board->cells[x][y].type == (dvonn_cell_type_t) board->player))
		continue;
	    height = board->cells[x][y].height;
	    if (dvonn_board_move_legal (board, x, y,
					x + height, y, &error))
		return TRUE;
	    if (dvonn_board_move_legal (board, x, y,
					x - height, y, &error))
		return TRUE;
	    if (dvonn_board_move_legal (board, x, y,
					x, y + height, &error))
		return TRUE;
	    if (dvonn_board_move_legal (board, x, y,
					x, y - height, &error))
		return TRUE;
	    if (dvonn_board_move_legal (board, x, y,
					x + height, y - height, &error))
		return TRUE;
	    if (dvonn_board_move_legal (board, x, y,
					x - height, y + height, &error))
		return TRUE;
	}
    }
    return FALSE;
}


static dvonn_phase_t
dvonn_board_next_player (dvonn_board_t *board)
{
    if (board->player == DVONN_PLAYER_BLACK)
	board->player = DVONN_PLAYER_WHITE;
    else
	board->player = DVONN_PLAYER_BLACK;

    /* Only look for legal moves during the movement phase. */
    if (board->phase != DVONN_PHASE_MOVEMENT)
	return board->phase;

    if (! dvonn_board_player_has_legal_move (board)) {
	if (board->player == DVONN_PLAYER_BLACK)
	    board->player = DVONN_PLAYER_WHITE;
	else
	    board->player = DVONN_PLAYER_BLACK;

	if (! dvonn_board_player_has_legal_move (board))
	{
	    int x, y;
	    dvonn_cell_t *cell;

	    board->phase = DVONN_PHASE_GAME_OVER;

	    /* Compute final score. */
	    for (x = 0; x < DVONN_BOARD_X_SIZE; x++) {
		for (y = 0; y < DVONN_BOARD_Y_SIZE; y++) {
		    if (dvonn_board_cell_occupied (board, x, y))
		    {
			cell = &board->cells[x][y];
			if (cell->type == DVONN_CELL_RED)
			    continue;
			board->score[cell->type] += cell->height;
		    }
		}
	    }
	}
    }

    return board->phase;
}

int
dvonn_board_place (dvonn_board_t *board,
		   int x, int y,
		   char **error)
{
    if (board->phase != DVONN_PHASE_PLACEMENT) {
	*error = "Cannot place outside of placement phase";
	return FALSE;
    }

    if (board->cells[x][y].type != DVONN_CELL_EMPTY) {
	*error = "Cannot place on an occupied space";
	return FALSE;
    }

    if (board->moves < 3) {
	board->cells[x][y].type = DVONN_CELL_RED;
	board->cells[x][y].contains_red = TRUE;
    } else if (board->moves % 2) {
	board->cells[x][y].type = DVONN_CELL_BLACK;
    } else {
	board->cells[x][y].type = DVONN_CELL_WHITE;
    }

    board->cells[x][y].height = 1;

    board->moves++;

    dvonn_board_next_player (board);

    if (board->moves == 49) {
	board->phase = DVONN_PHASE_MOVEMENT;
	board->moves = 0;
	board->player = DVONN_PLAYER_WHITE;
    }

    return TRUE;
}

static void
fill_living(dvonn_board_t *board,
	    int living[DVONN_BOARD_X_SIZE][DVONN_BOARD_Y_SIZE],
	    int x, int y)
{
    if (dvonn_board_cell_occupied (board, x, y) && ! living[x][y])
    {
	living[x][y] = 1;

	fill_living (board, living, x - 1, y);
	fill_living (board, living, x + 1, y);
	fill_living (board, living, x, y - 1);
	fill_living (board, living, x, y + 1);
	fill_living (board, living, x + 1, y - 1);
	fill_living (board, living, x - 1, y + 1);
    }
}

static void
eliminate_disconnected_stacks (dvonn_board_t *board)
{
    int x, y;
    int living[DVONN_BOARD_X_SIZE][DVONN_BOARD_Y_SIZE];

    for (x = 0; x < DVONN_BOARD_X_SIZE; x++)
	for (y = 0; y < DVONN_BOARD_Y_SIZE; y++)
	    living[x][y] = 0;

    for (x = 0; x < DVONN_BOARD_X_SIZE; x++)
	for (y = 0; y < DVONN_BOARD_Y_SIZE; y++)
	    if (board->cells[x][y].contains_red)
		fill_living (board, living, x, y);

    for (x = 0; x < DVONN_BOARD_X_SIZE; x++)
	for (y = 0; y < DVONN_BOARD_Y_SIZE; y++)
	    if (dvonn_board_cell_occupied (board, x, y) &&
		! living[x][y])
	    {
		board->cells[x][y].type = DVONN_CELL_EMPTY;
		board->cells[x][y].height = 0;
	    }
}

int
dvonn_board_move (dvonn_board_t *board,
		  int x1, int y1,
		  int x2, int y2,
		  char **error)
{
    if (board->phase != DVONN_PHASE_MOVEMENT) {
	*error = "Cannot move outside of placement phase";
	return FALSE;
    }

    if (! dvonn_board_move_legal (board, x1, y1, x2, y2, error))
	return FALSE;

    board->cells[x2][y2].height += board->cells[x1][y1].height;
    board->cells[x2][y2].type = board->cells[x1][y1].type;
    board->cells[x2][y2].contains_red |= board->cells[x1][y1].contains_red;

    board->cells[x1][y1].type = DVONN_CELL_EMPTY;
    board->cells[x1][y1].height = 0;
    board->cells[x1][y1].contains_red = FALSE;

    eliminate_disconnected_stacks (board);

    dvonn_board_next_player (board);

    return TRUE;
}
